/*
  NanodeRF_multinode

  Relays data received from wireless nodes to emoncms
  Decodes reply from server to set software real time clock
  Relays time data to emonglcd - and any other listening nodes.
  Looks for 'ok' reply from request to verify data reached emoncms

  emonBase Documentation: http://openenergymonitor.org/emon/emonbase

  Authors: Trystan Lea and Glyn Hudson
  Part of the: openenergymonitor.org project
  Licenced under GNU GPL V3
  http://openenergymonitor.org/emon/license

  EtherCard Library by Jean-Claude Wippler and Andrew Lindsay
  JeeLib Library by Jean-Claude Wippler

  THIS SKETCH REQUIRES:+
  
  Libraries in the standard arduino libraries folder:
	- JeeLib		https://github.com/jcw/jeelib
	- EtherCard		https://github.com/jcw/ethercard/

  Other files in project directory (should appear in the arduino tabs above)
	- decode_reply.ino
	- dhcp_dns.ino
*/

#define UNO       //anti crash wachdog reset only works with Uno (optiboot) bootloader, comment out the line if using delianuova

#define RF69_COMPAT 0
#include <JeeLib.h>	     //https://github.com/jcw/jeelib
#include <avr/wdt.h>

// for test
//#include <MemoryFree.h>

#define MYNODE 30            
#define RF_freq RF12_868MHZ     // RF_frequency
#define group 210            // network group

// Settings for various host hardware

#define SS 8             // for use with NanodeRF, or 10 for use with Open Kontrol Gateway 

#define RF12_CS 10       // 10 for emonTx, emonGLCD, NanodeRF, JeeNode; 9 for Open Kontrol Gateway


//---------------------------------------------------------------------
// The PacketBuffer class is used to generate the json string that is send via ethernet - JeeLabs
//---------------------------------------------------------------------
class PacketBuffer : public Print {
public:
    PacketBuffer () : fill (0) {}
    const char* buffer() { return buf; }
    byte length() { return fill; }
    void reset()
    { 
      memset(buf,0,sizeof(buf));
      fill = 0; 
    }
    virtual size_t write (uint8_t ch)
        { if (fill < sizeof buf) buf[fill++] = ch; }
    byte fill;
    char buf[250];
    private:
};
PacketBuffer str;

//--------------------------------------------------------------------------
// Ethernet
//--------------------------------------------------------------------------
#include <EtherCard.h>		//https://github.com/jcw/ethercard 



// ethernet interface mac address, must be unique on the LAN
static byte mymac[] = { 0x42,0x31,0x42,0x21,0x30,0x31 };

// 1) Set this to the domain name of your hosted emoncms - leave blank if posting to IP address 
const char website[] PROGMEM = "emoncms.org";

// or if your posting to a static IP server:
static byte hisip[] = { 192,168,1,10 };
// change to true if you would like the sketch to use hisip
boolean use_hisip = false;  



// 2) If your emoncms install is in a subdirectory add details here i.e "/emoncms3"
char basedir[] = "";

// 3) Set to your account write apikey 
char apikey[] = "feeddeadbeefbeadfeeddeadbeefbead";

//IP address of remote server, only needed when posting to a server that has not got a dns domain name (staticIP e.g local server) 
byte Ethernet::buffer[700];
static uint32_t timer;

const int redLED = 6;                     // NanodeRF RED indicator LED
//const int redLED = 17;  		          // Open Kontrol Gateway LED indicator
const int greenLED = 5;                   // NanodeRF GREEN indicator LED

int ethernet_error = 0;                   // Etherent (controller/DHCP) error flag
int rf_error = 0;                         // RF error flag - high when no data received 
int ethernet_requests = 0;                // count ethernet requests without reply                 

int dhcp_status = 0;
int dns_status = 0;

int data_ready=0;                         // Used to signal that emontx data is ready to be sent
unsigned long last_rf;                    // Used to check for regular emontx data - otherwise error

char line_buf[50];                        // Used to store line of http reply header

unsigned long time60s = -50000;
//**********************************************************************************************************************
// SETUP
//**********************************************************************************************************************
void setup () {
  
  //Nanode RF LED indicator  setup - green flashing means good - green or red on for a long time means bad! 
  //High means off since NanodeRF tri-state buffer inverts signal 
  pinMode(redLED, OUTPUT); digitalWrite(redLED,LOW);            
  pinMode(greenLED, OUTPUT); digitalWrite(greenLED,LOW);       
  delay(100); digitalWrite(redLED,HIGH);                          // turn off redLED
  
  Serial.begin(9600);
  Serial.println(F("\n[webClient]"));

  dhcp_status = 0;
  dns_status = 0;
  ethernet_requests = 0;
  ethernet_error=0;
  rf_error=0;

  if (ether.begin(sizeof Ethernet::buffer, mymac, SS) == 0) 
  {
    Serial.println(F( "Failed to access Ethernet controller"));
    ethernet_error = 1;  
  }

  rf12_set_cs(RF12_CS); 

  rf12_initialize(MYNODE, RF_freq, group);
  last_rf = millis()-40000;                                       // setting lastRF back 40s is useful as it forces the ethernet code to run straight away
   
  digitalWrite(greenLED,HIGH);                                    // Green LED off - indicate that setup has finished 
 
  #ifdef UNO
  wdt_enable(WDTO_8S); 
  #endif
}

//**********************************************************************************************************************
// LOOP
//**********************************************************************************************************************
void loop () {
  
  #ifdef UNO
  wdt_reset();
  #endif

  dhcp_dns();   // handle dhcp and dns setup - see dhcp_dns tab
  
  // Display error states on status LED
  if (ethernet_error==1 || rf_error==1 || ethernet_requests > 0) digitalWrite(redLED,LOW);
    else digitalWrite(redLED,HIGH);

  //-----------------------------------------------------------------------------------------------------------------
  // 1) On RF receive
  //-----------------------------------------------------------------------------------------------------------------
  if (rf12_recvDone()){      
      if (rf12_crc == 0 && (rf12_hdr & RF12_HDR_CTL) == 0)
      {
        int node_id = (rf12_hdr & 0x1F);
        byte n = rf12_len;
         
        str.reset();
        str.print(basedir); str.print(F("/input/post.json?"));
        str.print(F("apikey=")); str.print(apikey);
        str.print(F("&node="));  str.print(node_id);
        str.print(F("&csv="));
        for (byte i=0; i<n; i+=2)
        {
          int num = ((unsigned char)rf12_data[i+1] << 8 | (unsigned char)rf12_data[i]);
          if (i) str.print(",");
          str.print(num);
        }

        str.print("\0");  //  End of json string
        data_ready = 1; 
        last_rf = millis(); 
        rf_error=0;
      }
  }

  //-----------------------------------------------------------------------------------------------------------------
  // 2) If no data is received from rf12 module the server is updated every 30s with RFfail = 1 indicator for debugging
  //-----------------------------------------------------------------------------------------------------------------
  if ((millis()-last_rf)>30000)
  {
    last_rf = millis();                                                 // reset lastRF timer
    str.reset();                                                        // reset json string
    str.print(basedir); str.print(F("/input/post.json?"));
    str.print(F("apikey="));str.print(apikey);
    str.print(F("&json={rf_fail:1}\0"));                                // No RF received in 30 seconds so send failure 
    data_ready = 1;                                                     // Ok, data is ready
    rf_error=1;
  }

  //-----------------------------------------------------------------------------------------------------------------
  // 3) Send data via ethernet
  //-----------------------------------------------------------------------------------------------------------------
  ether.packetLoop(ether.packetReceive());
  
  if (data_ready) {
    
    Serial.print(F("Data sent: ")); Serial.println(str.buf); // print to serial json string

    // Example of posting to emoncms.org http://emoncms.org 
    // To point to your account just enter your WRITE APIKEY 
    ethernet_requests ++;
    ether.browseUrl(PSTR("") ,str.buf, website, my_callback);
//    ether.browseUrl(PSTR("/input/post.json?apikey=") ,str.buf, website, my_callback);
    data_ready =0;
  }
  
  if (ethernet_requests > 10) delay(10000); // Reset the nanode if more than 10 request attempts have been tried without a reply

  if ((millis()-time60s)>60000)
  {
    time60s = millis();                                                 // reset lastRF timer
    str.reset();
    str.print(basedir); str.print(F("/time/local.json?")); str.print(F("apikey=")); str.print(apikey);
    Serial.println(F("Time request sent"));
    ether.browseUrl(PSTR("") ,str.buf, website, my_callback);
  }
// for testing
//    Serial.print(F("freeMemory()="));
//    Serial.println(freeMemory());

delay(8);  
}
//**********************************************************************************************************************

//-----------------------------------------------------------------------------------
// Ethernet callback
// receive reply and decode
//-----------------------------------------------------------------------------------
static void my_callback (byte status, word off, word len) {
  int lsize =   get_reply_data(off);

  if (strcmp(line_buf,"ok")==0)
  {
    Serial.println(F("OK received")); ethernet_requests = 0; ethernet_error = 0;
  }
  else if(line_buf[0]=='t')
  {
    Serial.print("Time: ");
    Serial.println(line_buf);
    
/*    char tmp[] = {line_buf[1],line_buf[2],0};
    byte hour = atoi(tmp);
    tmp[0] = line_buf[4]; tmp[1] = line_buf[5];
    byte minute = atoi(tmp);
    tmp[0] = line_buf[7]; tmp[1] = line_buf[8];
    byte second = atoi(tmp);
*/
	int j = (line_buf[3]==',') ? 1: 0;
    byte hour = (atoi(line_buf+1) + 24) % 24;  //make hour non-negative
    byte minute = atoi(line_buf+3+j);
    byte second = atoi(line_buf+6+j);

    if (hour>0 || minute>0 || second>0) 
    {  
      char data[] = {'t',(char)hour,(char)minute,(char)second};

      int i = 0; while (!rf12_canSend() && i<10) {rf12_recvDone(); i++;}
      rf12_sendStart(0, data, sizeof data);
      rf12_sendWait(0);
/*
	  rf12_sendNow(0, data, sizeof data);
      rf12_sendWait(2);
*/	  
    }
	delay(2);
  }
}
